from __future__ import annotations

from functools import reduce
from typing import Any
from typing import Callable
from typing import Iterable

from narwhals.pandas_like import dtypes
from narwhals.pandas_like.dataframe import DataFrame
from narwhals.pandas_like.expr import Expr
from narwhals.pandas_like.series import Series
from narwhals.pandas_like.utils import horizontal_concat
from narwhals.pandas_like.utils import parse_into_exprs
from narwhals.pandas_like.utils import series_from_iterable
from narwhals.spec import IntoExpr
from narwhals.spec import Namespace as NamespaceProtocol
from narwhals.utils import flatten_str


class Namespace(NamespaceProtocol):
    Int64 = dtypes.Int64
    Int32 = dtypes.Int32
    Int16 = dtypes.Int16
    Int8 = dtypes.Int8
    UInt64 = dtypes.UInt64
    UInt32 = dtypes.UInt32
    UInt16 = dtypes.UInt16
    UInt8 = dtypes.UInt8
    Float64 = dtypes.Float64
    Float32 = dtypes.Float32
    Bool = dtypes.Bool
    String = dtypes.String

    # --- not in spec ---
    def __init__(self, implementation: str) -> None:
        self._implementation = implementation

    def _create_expr_from_callable(  # noqa: PLR0913
        self,
        func: Callable[[DataFrame], list[Series]],
        *,
        depth: int,
        function_name: str,
        root_names: list[str] | None,
        output_names: list[str] | None,
    ) -> Expr:
        return Expr(
            func,
            depth=depth,
            function_name=function_name,
            root_names=root_names,
            output_names=output_names,
            implementation=self._implementation,
        )

    def _create_series_from_scalar(self, value: Any, series: Series) -> Series:
        return Series(
            series_from_iterable(
                [value],
                name=series.series.name,
                index=series.series.index[0:1],
                implementation=self._implementation,
            ),
            implementation=self._implementation,
        )

    def _create_expr_from_series(self, series: Series) -> Expr:
        return Expr(
            lambda _df: [series],
            depth=0,
            function_name="series",
            root_names=None,
            output_names=None,
            implementation=self._implementation,
        )

    # --- selection ---
    def col(self, *column_names: str | Iterable[str]) -> Expr:
        return Expr.from_column_names(
            *flatten_str(*column_names), implementation=self._implementation
        )

    def all(self) -> Expr:
        return Expr(
            lambda df: [
                Series(
                    df._dataframe.loc[:, column_name],
                    implementation=self._implementation,
                )
                for column_name in df.columns
            ],
            depth=0,
            function_name="all",
            root_names=None,
            output_names=None,
            implementation=self._implementation,
        )

    # --- reduction ---
    def sum(self, *column_names: str) -> Expr:
        return Expr.from_column_names(
            *column_names, implementation=self._implementation
        ).sum()

    def mean(self, *column_names: str) -> Expr:
        return Expr.from_column_names(
            *column_names, implementation=self._implementation
        ).mean()

    def max(self, *column_names: str) -> Expr:
        return Expr.from_column_names(
            *column_names, implementation=self._implementation
        ).max()

    def min(self, *column_names: str) -> Expr:
        return Expr.from_column_names(
            *column_names, implementation=self._implementation
        ).min()

    def len(self) -> Expr:
        return Expr(
            lambda df: [
                Series(
                    series_from_iterable(
                        [len(df._dataframe)],
                        name="len",
                        index=[0],
                        implementation=self._implementation,
                    ),
                    implementation=self._implementation,
                ),
            ],
            depth=0,
            function_name="len",
            root_names=None,
            output_names=["len"],
            implementation=self._implementation,
        )

    # --- horizontal ---
    def sum_horizontal(self, *exprs: IntoExpr | Iterable[IntoExpr]) -> Expr:
        return reduce(lambda x, y: x + y, parse_into_exprs(self._implementation, *exprs))

    def all_horizontal(self, *exprs: IntoExpr | Iterable[IntoExpr]) -> Expr:
        return reduce(lambda x, y: x & y, parse_into_exprs(self._implementation, *exprs))

    def any_horizontal(self, *exprs: IntoExpr | Iterable[IntoExpr]) -> Expr:
        return reduce(lambda x, y: x | y, parse_into_exprs(self._implementation, *exprs))

    def concat(
        self,
        items: Iterable[DataFrame],  # type: ignore[override]
        *,
        how: str,
    ) -> DataFrame:
        dfs: list[Any] = []
        kind: Any = {}
        for df in items:
            dfs.append(df._dataframe)
            kind.append(type(df._dataframe))
        if len(kind) > 1:
            msg = "Can only concat DataFrames or LazyFrames, not mixtures of the two"
            raise TypeError(msg)
        if how != "horizontal":
            msg = "Only horizontal concatenation is supported for now"
            raise TypeError(msg)
        return DataFrame(
            horizontal_concat(dfs, implementation=self._implementation),
            implementation=self._implementation,
            # TODO (incorrect!)
            is_eager=True,
            is_lazy=False,
        )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Iterable
from typing import Literal
from typing import Sequence

import polars as pl

from narwhals.spec import DataFrame as DataFrameProtocol
from narwhals.spec import DType as DTypeProtocol
from narwhals.spec import Expr as ExprProtocol
from narwhals.spec import ExprStringNamespace as ExprStringNamespaceProtocol
from narwhals.spec import GroupBy as GroupByProtocol
from narwhals.spec import Namespace as NamespaceProtocol
from narwhals.spec import Series as SeriesProtocol
from narwhals.utils import flatten_into_expr

if TYPE_CHECKING:
    from typing_extensions import Self

    from narwhals.spec import IntoExpr


def extract_native(obj: Any) -> Any:
    if isinstance(obj, Expr):
        return obj._expr
    if isinstance(obj, DType):
        return obj._dtype
    if isinstance(obj, DataFrame):
        return obj._dataframe
    if isinstance(obj, Series):
        return obj._series
    return obj


class Expr(ExprProtocol):
    def __init__(self, expr: pl.Expr) -> None:
        self._expr = expr

    # --- convert ---
    def alias(self, name: str) -> Self:
        return self.__class__(self._expr.alias(name))

    def cast(
        self,
        dtype: DType,  # type: ignore[override]
    ) -> Self:
        return self.__class__(self._expr.cast(extract_native(dtype)))

    # --- binary ---
    def __eq__(self, other: object) -> Expr:  # type: ignore[override]
        return self.__class__(self._expr.__eq__(extract_native(other)))

    def __and__(self, other: Any) -> Expr:
        return self.__class__(self._expr.__and__(extract_native(other)))

    def __or__(self, other: Any) -> Expr:
        return self.__class__(self._expr.__or__(extract_native(other)))

    def __add__(self, other: Any) -> Expr:
        return self.__class__(self._expr.__add__(extract_native(other)))

    def __radd__(self, other: Any) -> Expr:
        return self.__class__(self._expr.__radd__(extract_native(other)))

    def __sub__(self, other: Any) -> Expr:
        return self.__class__(self._expr.__sub__(extract_native(other)))

    def __rsub__(self, other: Any) -> Expr:
        return self.__class__(self._expr.__rsub__(extract_native(other)))

    def __mul__(self, other: Any) -> Expr:
        return self.__class__(self._expr.__mul__(extract_native(other)))

    def __rmul__(self, other: Any) -> Expr:
        return self.__class__(self._expr.__rmul__(extract_native(other)))

    def __le__(self, other: Any) -> Expr:
        return self.__class__(self._expr.__le__(extract_native(other)))

    def __lt__(self, other: Any) -> Expr:
        return self.__class__(self._expr.__lt__(extract_native(other)))

    def __gt__(self, other: Any) -> Expr:
        return self.__class__(self._expr.__gt__(extract_native(other)))

    def __ge__(self, other: Any) -> Expr:
        return self.__class__(self._expr.__ge__(extract_native(other)))

    # --- unary ---
    def mean(self) -> Expr:
        return self.__class__(self._expr.mean())

    def sum(self) -> Expr:
        return self.__class__(self._expr.sum())

    def min(self) -> Expr:
        return self.__class__(self._expr.min())

    def max(self) -> Expr:
        return self.__class__(self._expr.max())

    def n_unique(self) -> Expr:
        return self.__class__(self._expr.n_unique())

    # --- transform ---
    def is_between(
        self, lower_bound: Any, upper_bound: Any, closed: str = "both"
    ) -> Expr:
        return self.__class__(self._expr.is_between(lower_bound, upper_bound, closed))  # type: ignore[arg-type]

    def is_in(self, other: Any) -> Expr:
        return self.__class__(self._expr.is_in(other))

    def is_null(self) -> Expr:
        return self.__class__(self._expr.is_null())

    # --- partial reduction ---
    def drop_nulls(self) -> Expr:
        return self.__class__(self._expr.drop_nulls())

    def sample(self, n: int, fraction: float, *, with_replacement: bool) -> Expr:
        return self.__class__(
            self._expr.sample(n, fraction=fraction, with_replacement=with_replacement)
        )

    # --- namespaces ---
    @property
    def str(self) -> ExprStringNamespace:
        return ExprStringNamespace(self._expr.str)


class ExprStringNamespace(ExprStringNamespaceProtocol):
    def __init__(self, expr: Any) -> None:
        self._expr = expr

    def ends_with(self, suffix: str) -> Expr:
        return Expr(self._expr.str.ends_with(suffix))


class DType(DTypeProtocol):
    def __init__(self, dtype: Any) -> None:
        self._dtype = dtype

    @classmethod
    def is_numeric(cls: type[Self]) -> bool:
        return cls._dtype.is_numeric()  # type: ignore[no-any-return]


class Namespace(NamespaceProtocol):
    Float64: DType
    Float32: DType
    Int64: DType
    Int32: DType
    Int16: DType
    Int8: DType
    UInt64: DType
    UInt32: DType
    UInt16: DType
    UInt8: DType
    Bool: DType
    String: DType

    # --- selection ---
    def col(self, *names: str | Iterable[str]) -> Expr:
        return Expr(pl.col(*names))  # type: ignore[arg-type]

    def all(self) -> Expr:
        return Expr(pl.all())

    # --- reduction ---
    def sum(self, *columns: str) -> Expr:
        return Expr(pl.sum(*columns))

    def mean(self, *columns: str) -> Expr:
        return Expr(pl.mean(*columns))

    def max(self, *columns: str) -> Expr:
        return Expr(pl.max(*columns))

    def min(self, *columns: str) -> Expr:
        return Expr(pl.min(*columns))

    def len(self) -> Expr:
        return Expr(pl.len())

    # --- horizontal ---
    def all_horizontal(self, *exprs: IntoExpr | Iterable[IntoExpr]) -> Expr:
        return Expr(pl.all_horizontal(*[extract_native(v) for v in exprs]))

    def any_horizontal(self, *exprs: IntoExpr | Iterable[IntoExpr]) -> Expr:
        return Expr(pl.any_horizontal(*[extract_native(v) for v in exprs]))

    def sum_horizontal(self, *exprs: IntoExpr | Iterable[IntoExpr]) -> Expr:
        return Expr(pl.sum_horizontal(*[extract_native(v) for v in exprs]))

    def concat(
        self,
        items: Iterable[DataFrame],  # type: ignore[override]
        *,
        how: str,
    ) -> DataFrame:
        # bit harder, do this later
        raise NotImplementedError


class Series(SeriesProtocol):
    def __init__(self, series: pl.Series) -> None:
        self._series = series

    def alias(self, name: str) -> Self:
        return self.__class__(self._series.alias(name))

    @property
    def name(self) -> str:
        return self._series.name

    @property
    def shape(self) -> tuple[int]:
        return self._series.shape

    def cast(
        self,
        dtype: DType,  # type: ignore[override]
    ) -> Self:
        return self.__class__(self._series.cast(extract_native(dtype)))

    def item(self) -> Any:
        return self._series.item()

    def is_between(
        self, lower_bound: Any, upper_bound: Any, closed: str = "both"
    ) -> Series:
        return self.__class__(self._series.is_between(lower_bound, upper_bound, closed))  # type: ignore[arg-type]

    def is_in(self, other: Any) -> Series:
        return self.__class__(self._series.is_in(other))

    def is_null(self) -> Series:
        return self.__class__(self._series.is_null())

    def drop_nulls(self) -> Series:
        return self.__class__(self._series.drop_nulls())

    def n_unique(self) -> int:
        return self._series.n_unique()

    def zip_with(self, mask: Self, other: Self) -> Self:
        return self.__class__(
            self._series.zip_with(extract_native(mask), extract_native(other))
        )

    def sample(self, n: int, fraction: float, *, with_replacement: bool) -> Series:
        return self.__class__(
            self._series.sample(n, fraction=fraction, with_replacement=with_replacement)
        )

    def to_numpy(self) -> Any:
        return self._series.to_numpy()

    def to_pandas(self) -> Any:
        return self._series.to_pandas()


class DataFrame(DataFrameProtocol):
    def __init__(
        self, df: pl.DataFrame | pl.LazyFrame, *, is_eager: bool, is_lazy: bool
    ) -> None:
        self._dataframe = df
        self._is_eager = is_eager
        self._is_lazy = is_lazy

    def _from_dataframe(self, df: pl.DataFrame | pl.LazyFrame) -> Self:
        # construct, preserving properties
        return self.__class__(df, is_eager=self._is_eager, is_lazy=self._is_lazy)

    def __getitem__(self, column_name: str) -> Series:
        if not self._is_eager:
            raise RuntimeError(
                "DataFrame.shape can only be called if frame was instantiated with `is_eager=True`"
            )
        assert isinstance(self._dataframe, pl.DataFrame)
        return Series(
            self._dataframe[column_name],
        )

    # --- properties ---
    @property
    def columns(self) -> list[str]:
        return self._dataframe.columns

    @property
    def schema(self) -> dict[str, DTypeProtocol]:
        return {key: DType(value) for key, value in self._dataframe.schema.items()}

    @property
    def shape(self) -> tuple[int, int]:
        if not self._is_eager:
            raise RuntimeError(
                "DataFrame.shape can only be called if frame was instantiated with `is_eager=True`"
            )
        assert isinstance(self._dataframe, pl.DataFrame)
        return self._dataframe.shape

    # --- reshape ---
    def with_columns(
        self, *exprs: IntoExpr | Iterable[IntoExpr], **named_exprs: IntoExpr
    ) -> Self:
        return self._from_dataframe(
            self._dataframe.with_columns(
                *[extract_native(v) for v in exprs],
                **{key: extract_native(value) for key, value in named_exprs.items()},
            )
        )

    def filter(self, *predicates: IntoExpr | Iterable[IntoExpr]) -> Self:
        return self._from_dataframe(
            self._dataframe.filter(*[extract_native(v) for v in predicates])
        )

    def select(
        self, *exprs: IntoExpr | Iterable[IntoExpr], **named_exprs: IntoExpr
    ) -> Self:
        return self._from_dataframe(
            self._dataframe.select(
                *[extract_native(v) for v in exprs],
                **{key: extract_native(value) for key, value in named_exprs.items()},
            )
        )

    def rename(self, mapping: dict[str, str]) -> Self:
        return self._from_dataframe(self._dataframe.rename(mapping))

    # --- transform ---
    def sort(
        self,
        by: str | Iterable[str],
        *more_by: str,
        descending: bool | Sequence[bool] = False,
    ) -> Self:
        return self._from_dataframe(
            self._dataframe.sort(by, *more_by, descending=descending)
        )

    # --- convert ---
    def lazy(self) -> Self:
        return self.__class__(self._dataframe.lazy(), is_eager=False, is_lazy=True)

    def collect(self) -> Self:
        if not self._is_lazy:
            raise RuntimeError(
                "DataFrame.collect can only be called if frame was instantiated with `is_lazy=True`"
            )
        assert isinstance(self._dataframe, pl.LazyFrame)
        return self.__class__(self._dataframe.collect(), is_eager=True, is_lazy=False)

    def cache(self) -> Self:
        if not self._is_lazy:
            raise RuntimeError(
                "DataFrame.cache can only be called if frame was instantiated with `is_lazy=True`"
            )
        assert isinstance(self._dataframe, pl.LazyFrame)
        return self.__class__(self._dataframe.cache(), is_eager=False, is_lazy=True)

    def to_numpy(self) -> Any:
        if not self._is_eager:
            raise RuntimeError(
                "DataFrame.to_numpy can only be called if frame was instantiated with `is_eager=True`"
            )
        assert isinstance(self._dataframe, pl.DataFrame)
        return self._dataframe.to_numpy()

    def to_pandas(self) -> Any:
        if not self._is_eager:
            raise RuntimeError(
                "DataFrame.to_pandas can only be called if frame was instantiated with `is_eager=True`"
            )
        assert isinstance(self._dataframe, pl.DataFrame)
        return self._dataframe.to_pandas()

    def to_dict(self, *, as_series: bool = True) -> dict[str, Any]:
        if not self._is_eager:
            raise RuntimeError(
                "DataFrame.to_dict can only be called if frame was instantiated with `is_eager=True`"
            )
        assert isinstance(self._dataframe, pl.DataFrame)
        return self._dataframe.to_dict(as_series=as_series)

    # --- actions ---
    def join(
        self,
        other: Self,
        *,
        how: Literal["inner"] = "inner",
        left_on: str | list[str],
        right_on: str | list[str],
    ) -> Self:
        # todo validate eager/lazy only
        return self._from_dataframe(
            self._dataframe.join(
                extract_native(other), how=how, left_on=left_on, right_on=right_on
            )
        )

    def group_by(self, *keys: str | Iterable[str]) -> GroupBy:
        return GroupBy(
            self._dataframe.group_by(*keys),
            is_eager=self._is_eager,
            is_lazy=self._is_lazy,
        )

    # --- partial reduction ---
    def head(self, n: int) -> Self:
        return self._from_dataframe(self._dataframe.head(n))

    def unique(self, subset: list[str]) -> Self:
        return self._from_dataframe(self._dataframe.unique(subset))

    # --- public, non-Polars ---
    def to_native(self) -> Any:
        return self._dataframe

    @property
    def is_eager(self) -> bool:
        return self._is_eager

    @property
    def is_lazy(self) -> bool:
        return self._is_lazy


class GroupBy(GroupByProtocol):
    def __init__(self, groupby: Any, *, is_eager: bool, is_lazy: bool) -> None:
        self._groupby = groupby
        self._is_eager = is_eager
        self._is_lazy = is_lazy

    def agg(
        self, *aggs: IntoExpr | Iterable[IntoExpr], **named_aggs: IntoExpr
    ) -> DataFrame:
        return DataFrame(
            self._groupby.agg(
                *[extract_native(v) for v in flatten_into_expr(*aggs)],
                **{key: extract_native(value) for key, value in named_aggs.items()},
            ),
            is_eager=self._is_eager,
            is_lazy=self._is_lazy,
        )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

if TYPE_CHECKING:
    from narwhals.pandas_like.dataframe import DataFrame
    from narwhals.pandas_like.namespace import Namespace
    from narwhals.pandas_like.series import Series


def translate_frame(
    df: Any,
    implementation: str,
    *,
    is_eager: bool,
    is_lazy: bool,
) -> tuple[DataFrame, Namespace]:
    from narwhals.pandas_like.dataframe import DataFrame
    from narwhals.pandas_like.namespace import Namespace

    df = DataFrame(
        df,
        implementation=implementation,
        is_eager=is_eager,
        is_lazy=is_lazy,
    )
    return df, Namespace(implementation=implementation)


def translate_series(
    series: Any,
    implementation: str,
) -> tuple[Series, Namespace]:
    from narwhals.pandas_like.namespace import Namespace
    from narwhals.pandas_like.series import Series

    return Series(series, implementation=implementation), Namespace(
        implementation=implementation
    )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Iterable
from typing import Literal
from typing import Protocol
from typing import Sequence

if TYPE_CHECKING:
    from typing_extensions import Self


class Expr(Protocol):
    # --- convert ---
    def alias(self, name: str) -> Self:
        ...

    def cast(self, dtype: DType) -> Self:
        ...

    # --- binary ---
    def __eq__(self, other: object) -> Expr:  # type: ignore[override]
        ...

    def __and__(self, other: Any) -> Expr:
        ...

    def __or__(self, other: Any) -> Expr:
        ...

    def __add__(self, other: Any) -> Expr:
        ...

    def __radd__(self, other: Any) -> Expr:
        ...

    def __sub__(self, other: Any) -> Expr:
        ...

    def __rsub__(self, other: Any) -> Expr:
        ...

    def __mul__(self, other: Any) -> Expr:
        ...

    def __rmul__(self, other: Any) -> Expr:
        ...

    def __le__(self, other: Any) -> Expr:
        ...

    def __lt__(self, other: Any) -> Expr:
        ...

    def __gt__(self, other: Any) -> Expr:
        ...

    def __ge__(self, other: Any) -> Expr:
        ...

    # --- unary ---
    def mean(self) -> Expr:
        ...

    def sum(self) -> Expr:
        ...

    def min(self) -> Expr:
        ...

    def max(self) -> Expr:
        ...

    def n_unique(self) -> Expr:
        ...

    # --- transform ---
    def is_between(
        self, lower_bound: Any, upper_bound: Any, closed: str = "both"
    ) -> Expr:
        ...

    def is_in(self, other: Any) -> Expr:
        ...

    def is_null(self) -> Expr:
        ...

    # --- partial reduction ---
    def drop_nulls(self) -> Expr:
        ...

    def sample(self, n: int, fraction: float, *, with_replacement: bool) -> Expr:
        ...

    # --- namespaces ---
    @property
    def str(self) -> ExprStringNamespace:
        ...


class ExprStringNamespace(Protocol):
    def ends_with(self, suffix: str) -> Expr:
        ...


class DType(Protocol):
    @classmethod
    def is_numeric(cls: type[Self]) -> bool:
        ...


class Namespace(Protocol):
    Float64: DType
    Float32: DType
    Int64: DType
    Int32: DType
    Int16: DType
    Int8: DType
    UInt64: DType
    UInt32: DType
    UInt16: DType
    UInt8: DType
    Bool: DType
    String: DType

    # --- selection ---
    def col(self, *names: str | Iterable[str]) -> Expr:
        ...

    def all(self) -> Expr:
        ...

    # --- reduction ---
    def sum(self, *columns: str) -> Expr:
        ...

    def mean(self, *columns: str) -> Expr:
        ...

    def max(self, *columns: str) -> Expr:
        ...

    def min(self, *columns: str) -> Expr:
        ...

    def len(self) -> Expr:
        ...

    # --- horizontal ---
    def all_horizontal(self, *exprs: IntoExpr | Iterable[IntoExpr]) -> Expr:
        ...

    def any_horizontal(self, *exprs: IntoExpr | Iterable[IntoExpr]) -> Expr:
        ...

    def sum_horizontal(self, *exprs: IntoExpr | Iterable[IntoExpr]) -> Expr:
        ...

    def concat(self, items: Iterable[DataFrame], *, how: str) -> DataFrame:
        ...


class Series(Protocol):
    def alias(self, name: str) -> Self:
        ...

    @property
    def name(self) -> str:
        ...

    @property
    def shape(self) -> tuple[int]:
        ...

    def cast(self, dtype: DType) -> Self:
        ...

    def item(self) -> Any:
        ...

    def is_between(
        self, lower_bound: Any, upper_bound: Any, closed: str = "both"
    ) -> Series:
        ...

    def is_in(self, other: Any) -> Series:
        ...

    def is_null(self) -> Series:
        ...

    def drop_nulls(self) -> Series:
        ...

    def n_unique(self) -> int:
        ...

    def zip_with(self, mask: Self, other: Self) -> Self:
        ...

    def sample(self, n: int, fraction: float, *, with_replacement: bool) -> Series:
        ...

    def to_numpy(self) -> Any:
        ...

    def to_pandas(self) -> Any:
        ...


class DataFrame(Protocol):
    def __getitem__(self, column_name: str) -> Series:
        ...

    # --- properties ---
    @property
    def columns(self) -> list[str]:
        ...

    @property
    def schema(self) -> dict[str, DType]:
        ...

    @property
    def shape(self) -> tuple[int, int]:
        ...

    # --- reshape ---
    def with_columns(
        self, *exprs: IntoExpr | Iterable[IntoExpr], **named_exprs: IntoExpr
    ) -> Self:
        ...

    def filter(self, *predicates: IntoExpr | Iterable[IntoExpr]) -> Self:
        ...

    def select(
        self, *exprs: IntoExpr | Iterable[IntoExpr], **named_exprs: IntoExpr
    ) -> Self:
        ...

    def rename(self, mapping: dict[str, str]) -> Self:
        ...

    # --- transform ---
    def sort(
        self,
        by: str | Iterable[str],
        *more_by: str,
        descending: bool | Sequence[bool] = False,
    ) -> Self:
        ...

    # --- convert ---
    def lazy(self) -> Self:
        ...

    def collect(self) -> Self:
        ...

    def cache(self) -> Self:
        ...

    def to_numpy(self) -> Any:
        ...

    def to_pandas(self) -> Any:
        ...

    def to_dict(self, *, as_series: bool = True) -> dict[str, Any]:
        ...

    # --- actions ---
    def join(
        self,
        other: Self,
        *,
        how: Literal["inner"] = "inner",
        left_on: str | list[str],
        right_on: str | list[str],
    ) -> Self:
        ...

    def group_by(self, *keys: str | Iterable[str]) -> GroupBy:
        ...

    # --- partial reduction ---
    def head(self, n: int) -> Self:
        ...

    def unique(self, subset: list[str]) -> Self:
        ...

    # --- public, non-Polars ---
    def to_native(self) -> Any:
        ...

    @property
    def is_eager(self) -> bool:
        ...

    @property
    def is_lazy(self) -> bool:
        ...


class GroupBy(Protocol):
    def agg(
        self, *aggs: IntoExpr | Iterable[IntoExpr], **named_aggs: IntoExpr
    ) -> DataFrame:
        ...


IntoExpr = Expr | str | int | float | Series



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Iterable

if TYPE_CHECKING:
    from narwhals.spec import IntoExpr


def remove_prefix(text: str, prefix: str) -> str:
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def flatten_into_expr(*args: IntoExpr | Iterable[IntoExpr]) -> list[IntoExpr]:
    out: list[IntoExpr] = []
    for arg in args:
        if isinstance(arg, (list, tuple)):
            out.extend(arg)
        else:
            out.append(arg)  # type: ignore[arg-type]
    return out


def flatten_str(*args: str | Iterable[str]) -> list[str]:
    out: list[str] = []
    for arg in args:
        if isinstance(arg, str):
            out.append(arg)
        else:
            for item in arg:
                if not isinstance(item, str):
                    msg = f"Expected str, got {type(item)}"
                    raise TypeError(msg)
                out.append(item)
    return out


def flatten_bool(*args: bool | Iterable[bool]) -> list[bool]:
    out: list[bool] = []
    for arg in args:
        if isinstance(arg, bool):
            out.append(arg)
        else:
            for item in arg:
                if not isinstance(item, bool):
                    msg = f"Expected str, got {type(item)}"
                    raise TypeError(msg)
                out.append(item)
    return out



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import NamedTuple

from narwhals.dependencies import get_cudf
from narwhals.dependencies import get_modin
from narwhals.dependencies import get_pandas
from narwhals.dependencies import get_polars

if TYPE_CHECKING:
    from narwhals.spec import DataFrame
    from narwhals.spec import Namespace
    from narwhals.spec import Series


class FrameTranslation(NamedTuple):
    frame: DataFrame
    namespace: Namespace


class SeriesTranslation(NamedTuple):
    series: Series
    namespace: Namespace


def translate_frame(
    df: Any,
    *,
    is_eager: bool = False,
    is_lazy: bool = False,
) -> FrameTranslation:
    if is_eager and is_lazy:
        msg = "Only one of `is_eager` and `is_lazy` can be True."
        raise ValueError(msg)

    if hasattr(df, "__narwhals_frame__"):
        return FrameTranslation(
            *df.__narwhals_frame__(is_eager=is_eager, is_lazy=is_lazy)
        )

    if (pl := get_polars()) is not None:
        if isinstance(df, pl.LazyFrame) and is_eager:
            msg = (
                "Expected DataFrame, got LazyFrame. Set `is_eager=False` if you "
                "function doesn't require eager execution, or collect your frame "
                "before passing it to this function."
            )
            raise TypeError(msg)
        if isinstance(df, pl.DataFrame) and is_lazy:
            msg = (
                "Expected LazyFrame, got DataFrame. Set `is_lazy=False` if you "
                "function doesn't doesn't need to use `.collect`, or make your frame "
                "before passing it to this function."
            )
            raise TypeError(msg)
        if isinstance(df, (pl.DataFrame, pl.LazyFrame)):
            from narwhals.polars import DataFrame
            from narwhals.polars import Namespace

            return FrameTranslation(
                DataFrame(df, is_eager=is_eager, is_lazy=is_lazy), Namespace()
            )

    if (pd := get_pandas()) is not None and isinstance(df, pd.DataFrame):
        from narwhals.pandas_like.translate import translate_frame

        return FrameTranslation(
            *translate_frame(
                df,
                implementation="pandas",
                is_eager=is_eager,
                is_lazy=is_lazy,
            )
        )

    if (cudf := get_cudf()) is not None and isinstance(df, cudf.DataFrame):
        from narwhals.pandas_like.translate import translate_frame

        return FrameTranslation(
            *translate_frame(
                df, implementation="cudf", is_eager=is_eager, is_lazy=is_lazy
            )
        )

    if (mpd := get_modin()) is not None and isinstance(df, mpd.DataFrame):
        from narwhals.pandas_like.translate import translate_frame

        return FrameTranslation(
            *translate_frame(
                df, implementation="modin", is_eager=is_eager, is_lazy=is_lazy
            )
        )

    msg = f"Could not translate DataFrame {type(df)}, please open a feature request."
    raise NotImplementedError(msg)


def translate_series(
    series: Any,
) -> SeriesTranslation:
    if hasattr(series, "__narwhals_series__"):
        return SeriesTranslation(*series.__narwhals_series__())

    if (pl := get_polars()) is not None and isinstance(series, pl.Series):
        from narwhals.polars import Namespace
        from narwhals.polars import Series

        return SeriesTranslation(Series(series), Namespace())

    if (pd := get_pandas()) is not None and isinstance(series, pd.Series):
        from narwhals.pandas_like.translate import translate_series

        return SeriesTranslation(
            *translate_series(
                series,
                implementation="pandas",
            )
        )

    if (cudf := get_cudf()) is not None and isinstance(series, cudf.Series):
        from narwhals.pandas_like.translate import translate_series

        return SeriesTranslation(*translate_series(series, implementation="cudf"))

    if (mpd := get_modin()) is not None and isinstance(series, mpd.Series):
        from narwhals.pandas_like.translate import translate_series

        return SeriesTranslation(*translate_series(series, implementation="modin"))

    msg = f"Could not translate {type(series)}, please open a feature request."
    raise NotImplementedError(msg)


def translate_any(obj: Any) -> tuple[Series | DataFrame, Namespace]:
    try:
        return translate_series(obj)
    except NotImplementedError:
        return translate_frame(obj, is_eager=True)



from __future__ import annotations

from typing import TYPE_CHECKING

from narwhals.pandas_like.utils import isinstance_or_issubclass
from narwhals.spec import DType as DTypeProtocol

if TYPE_CHECKING:
    from typing_extensions import Self


class DType(DTypeProtocol):
    def __repr__(self) -> str:
        return self.__class__.__qualname__

    @classmethod
    def is_numeric(cls: type[Self]) -> bool:
        return issubclass(cls, NumericType)

    def __eq__(self, other: DType | type[DType]) -> bool:  # type: ignore[override]
        return isinstance_or_issubclass(other, type(self))


class NumericType(DType):
    ...


class Int64(NumericType):
    ...


class Int32(NumericType):
    ...


class Int16(NumericType):
    ...


class Int8(NumericType):
    ...


class UInt64(NumericType):
    ...


class UInt32(NumericType):
    ...


class UInt16(NumericType):
    ...


class UInt8(NumericType):
    ...


class Float64(NumericType):
    ...


class Float32(NumericType):
    ...


class String(DType):
    ...


class Bool(DType):
    ...



from narwhals import containers
from narwhals.translate import translate_any
from narwhals.translate import translate_frame
from narwhals.translate import translate_series

__version__ = "0.2.0"

__all__ = ["translate_frame", "translate_series", "translate_any", "containers"]



# todo: simplify all of this!
try:
    import polars as pl
except ModuleNotFoundError:
    POLARS_AVAILABLE = False
    pl = object  # type: ignore[assignment]
else:
    POLARS_AVAILABLE = True
try:
    import pandas as pd
except ModuleNotFoundError:
    PANDAS_AVAILABLE = False
    pd = object
else:
    PANDAS_AVAILABLE = True
try:
    import cudf
except ModuleNotFoundError:
    CUDF_AVAILABLE = False
    cudf = object
else:
    CUDF_AVAILABLE = True
try:
    import modin.pandas as mpd
except ModuleNotFoundError:
    MODIN_AVAILABLE = False
    mpd = object
else:
    MODIN_AVAILABLE = True


from typing import Any


def is_dataframe(obj: Any) -> bool:
    if POLARS_AVAILABLE and isinstance(
        obj, (pl.DataFrame, pl.LazyFrame, pl.Expr, pl.Series)
    ):
        return isinstance(obj, pl.DataFrame)
    from narwhals.pandas_like.dataframe import DataFrame

    return isinstance(obj, DataFrame)


def is_expr(obj: Any) -> bool:
    if POLARS_AVAILABLE and isinstance(
        obj, (pl.DataFrame, pl.LazyFrame, pl.Expr, pl.Series)
    ):
        return isinstance(obj, pl.Expr)
    from narwhals.pandas_like.expr import Expr

    return isinstance(obj, Expr)


def is_series(obj: Any) -> bool:
    if POLARS_AVAILABLE and isinstance(
        obj, (pl.DataFrame, pl.LazyFrame, pl.Expr, pl.Series)
    ):
        return isinstance(obj, pl.Series)
    from narwhals.pandas_like.series import Series

    return isinstance(obj, Series)


def get_implementation(obj: Any) -> str:
    if POLARS_AVAILABLE and isinstance(
        obj, (pl.DataFrame, pl.LazyFrame, pl.Expr, pl.Series)
    ):
        return "polars"
    if PANDAS_AVAILABLE and isinstance(obj, (pd.DataFrame, pd.Series)):
        return "pandas"
    if CUDF_AVAILABLE and isinstance(obj, (cudf.DataFrame, cudf.Series)):
        return "cudf"
    if MODIN_AVAILABLE and isinstance(obj, mpd.DataFrame):
        return "modin"
    msg = f"Unknown implementation: {obj}"
    raise TypeError(msg)


def is_pandas(obj: Any) -> bool:
    return get_implementation(obj) == "pandas"


def is_polars(obj: Any) -> bool:
    return get_implementation(obj) == "polars"


def is_cudf(obj: Any) -> bool:
    return get_implementation(obj) == "cudf"


def is_modin(obj: Any) -> bool:
    return get_implementation(obj) == "modin"
